<?php

declare(strict_types=1);

namespace Drupal\entity_reference_field_autocomplete_filter\Plugin\Field\FieldWidget;

use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\EntityReferenceAutocompleteWidget;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelInterface;
use Drupal\Core\Utility\Error;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin for the 'entity_reference_filterable_autocomplete' widget.
 *
 * @FieldWidget(
 *   id = "entity_reference_filterable_autocomplete",
 *   label = @Translation("Filterable Autocomplete"),
 *   description = @Translation("A filterable autocomplete text field."),
 *   field_types = {
 *     "entity_reference"
 *   }
 * )
 */
class EntityReferenceFilterableAutocompleteWidget extends EntityReferenceAutocompleteWidget {

  /**
   * Constructs a new instance of the object.
   *
   * @param string $plugin_id
   *   The plugin ID.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *   The field definition.
   * @param array $settings
   *   An array of settings.
   * @param array $third_party_settings
   *   An array of third party settings.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager service.
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    array $third_party_settings,
    private readonly EntityTypeManagerInterface $entityTypeManager,
    private readonly LoggerChannelInterface $logger,
  ) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition): static {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('entity_type.manager'),
      $container->get('logger.channel.entity_reference_field_autocomplete_filter'),
    );
  }

  /**
   * {@inheritDoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state): array {
    $element = parent::formElement($items, $delta, $element, $form, $form_state);
    $fieldName = $this->fieldDefinition->getName();
    $referencedEntities = $items->referencedEntities();
    $thisEntity = $referencedEntities[$delta] ?? NULL;
    $thisEntityBundle = $thisEntity?->bundle();

    $input = $form_state->getUserInput();
    $selectedBundle = $input[$fieldName][$delta]['bundle'] ?? $thisEntityBundle;
    $enabledBundles = $this->getFieldSetting('handler_settings')['target_bundles'];

    if ($selectedBundle) {
      $element['target_id']['#selection_settings']['target_bundles'] = [$selectedBundle];
    }
    else {
      // Limit to the bundles defined in the field settings.
      $element['target_id']['#selection_settings']['target_bundles'] = $enabledBundles;
    }

    // Add the bundle select field with AJAX callback.
    $wrapperId = Html::getClass($fieldName) . '-target-id-wrapper-' . $delta;
    $element['bundle'] = [
      '#type' => 'select',
      '#options' => $this->bundleOptions($enabledBundles),
      '#weight' => -1,
      '#empty_option' => 'All bundles',
      '#default_value' => $thisEntityBundle,
      '#ajax' => [
        'callback' => [static::class, 'updateTargetBundles'],
        'wrapper' => $wrapperId,
      ],
      '#wrapper_attributes' => [
        'class' => ['entity-reference-filterable-autocomplete-bundle'],
      ],
    ];

    // Wrap the target_id field so that it can be updated via AJAX.
    // Also wrap both the target_id & bundle fields for styling purposes.
    $element['bundle']['#prefix'] = '<div class="entity-reference-filterable-autocomplete-wrapper">';
    $element['target_id']['#prefix'] = '<div id="' . $wrapperId . '">';
    $element['target_id']['#suffix'] = '</div></div>';

    $element['#attached']['library'][] = 'entity_reference_field_autocomplete_filter/theme';

    return $element;
  }

  /**
   * Retrieves an array of bundle options with their labels.
   *
   * @param array $enabledBundles
   *   An array of enabled bundle IDs.
   *
   * @return array
   *   An associative array of bundle IDs => labels.
   */
  private function bundleOptions(array $enabledBundles): array {
    $fieldTargetType = $this->getFieldSetting('target_type');
    $storage = NULL;

    // Try and load the storage for this entity type.
    if ($this->entityTypeManager->hasHandler($fieldTargetType, 'storage')) {
      try {
        $storage = $this->entityTypeManager->getHandler($fieldTargetType, 'storage');
      }
      catch (\Exception $e) {
        Error::logException($this->logger, $e);
      }
    }

    $options = [];
    foreach ($enabledBundles as $bundle) {
      // If possible get the label for this bundle.
      $label = $storage?->load($bundle)?->label() ?? $bundle;
      $options[$bundle] = $label;
    }

    return $options;
  }

  /**
   * Respond to update target bundles ajax request.
   *
   * @param array $form
   *   The form structure from which the target bundles are retrieved.
   * @param \Drupal\Core\Form\FormStateInterface $formState
   *   The form state object.
   *
   * @return array
   *   The updated target_id field.
   */
  public static function updateTargetBundles(array $form, FormStateInterface $formState): array {
    $triggeringElement = $formState->getTriggeringElement();
    [$fieldName, $delta] = $triggeringElement['#parents'];
    return $form[$fieldName]['widget'][$delta]['target_id'];
  }

}
